# Copyright 2023 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Union, Optional
from requests import Response
from pandas import Series, DataFrame, to_datetime, json_normalize  # type: ignore
from spgci.api_client import get_data, Paginator
from spgci.utilities import odata_list_to_filter, list_to_filter
from urllib.parse import urlencode, quote, parse_qs, urlparse
from datetime import date
from enum import Enum


class WorldRefineryData:
    """
    World Refinery Data.

    Includes
    --------
    ``RefTypes`` to use with the ``get_reference_data`` method.
    ``get_reference_data()`` to get the list of countries, owners, refineries, etc.. used in the World Refinery Database.
    ``get_capacity()`` to get refinery capacity changes.
    ``get_runs()`` to get annual crude runs for individual refineries.
    ``get_yields()`` to get the yields for individual refineries.
    ``get_outages()`` to get refinery outages.
    ``get_outage_alerts()`` to get intra-day refinery outages.
    ``get_ownership()`` to get refinery ownership.
    ``get_margins()`` to get margin data.
    """

    _endpoint = "odata/refinery-data/v2.2/"

    class RefTypes(Enum):
        """World Refinery Database Reference Data Type"""

        CapacityStatuses = "CapacityStatuses"
        Cities = "cities"
        Configurations = "configurations"
        Countries = "countries"
        MarginTypes = "marginTypes"
        Operators = "operators"
        OutageUnits = "outageUnits"
        Owners = "owners"
        PADDs = "PADDs"
        # ProcessUnitCategories = "ProcessUnitCategories"
        ProcessUnits = "ProcessUnits"
        Refineries = "Refineries"
        Regions = "Regions"
        States = "States"

    @staticmethod
    def _padd_to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = json_normalize(
            j["value"],
            meta=["Id", "Name"],
            record_path="States",
            record_prefix="State.",
        )

        return df

    @staticmethod
    def _ts_to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = json_normalize(j, record_path="value")

        # drop odata artifacts
        drop = df.columns[df.columns.str.contains("@odata")]
        df = df[df.columns.drop(drop)]  # type: ignore

        if len(df) > 0 and "start_date" in df.columns:
            df["start_date"] = to_datetime(df["start_date"], utc=True)  # type: ignore
        if len(df) > 0 and "end_date" in df.columns:
            df["end_date"] = to_datetime(df["end_date"], utc=True)  # type: ignore

        return df

    @staticmethod
    def _to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = json_normalize(j["value"])

        # duplicates due to $expand=*
        dupes = ["RefineryId", "OwnerId", "CapacityStatusId", "MarginTypeId"]
        df = df[df.columns.drop(dupes, "ignore")]  # type: ignore

        if len(df) > 0 and "ModifiedDate" in df.columns:
            df["ModifiedDate"] = to_datetime(df["ModifiedDate"], utc=True)  # type: ignore
        if len(df) > 0 and "Date" in df.columns:
            df["Date"] = to_datetime(df["Date"], utc=True)  # type: ignore

        return df

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()

        count: int = j["@odata.count"]
        page_size = parse_qs(urlparse(resp.url).query)["pageSize"]

        if not page_size:
            return Paginator(False, "$skip", 0)

        remainder = count % int(page_size[0])
        quotient = count // int(page_size[0])
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "$skip", total_pages)

        return Paginator(True, "$skip", total_pages, pg_type="odata")

    def get_capacity(
        self,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        quarter: Optional[Union[int, list[int], "Series[int]"]] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        owner: Optional[Union[str, list[str], "Series[str]"]] = None,
        capacity_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        capacity_status_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        process_unit: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch historical refinery capacity changes.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        quarter : Optional[Union[int, list[int], Series[int]]], optional
            filter by quarter, by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        owner : Optional[Union[str, list[str], Series[str]]], optional
            filter by Owner/Name, by default None
        capacity_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by capacityId, by default None
        capacity_status_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by capacityStatusId, by default None
        process_unit : Optional[Union[str, list[str], Series[str]]], optional
            filter by ProcessUnit/Name, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Country/Name, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Region/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_capacity(year=2008)

        **Using Lists**
        >>> ci.WorldRefineryData().get_capacity(year_gte=2022, process_unit=["Atmos Distillation", "Dist Hydrocracking"])

        **Using Series**
        >>> df = ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.Refineries)
        >>> ci.WorldRefineryData().get_capacity(year=2020, refinery_id=df["Id"][:3])
        """
        endpoint = "capacity"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("Year", year))
        filter_params.append(odata_list_to_filter("Quarter", quarter))
        filter_params.append(odata_list_to_filter("RefineryId", refinery_id))
        filter_params.append(odata_list_to_filter("Owner/Name", owner))
        filter_params.append(odata_list_to_filter("CapacityId", capacity_id))
        filter_params.append(
            odata_list_to_filter("CapacityStatusId", capacity_status_id)
        )
        filter_params.append(odata_list_to_filter("ProcessUnit/Name", process_unit))
        filter_params.append(odata_list_to_filter("Refinery/Country/Name", country))
        filter_params.append(odata_list_to_filter("Refinery/Region/Name", region))

        if year_gt:
            filter_params.append(f"year gt {year_gt}")
        if year_gte:
            filter_params.append(f"year ge {year_gte}")
        if year_lt:
            filter_params.append(f"year lt {year_lt}")
        if year_lte:
            filter_params.append(f"year le {year_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
            "$expand": "*",
        }

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._to_df,
            paginate_fn=self._paginate,
        )

    def get_runs(
        self,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        quarter: Optional[Union[int, list[int], "Series[int]"]] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        owner: Optional[Union[str, list[str], "Series[int]"]] = None,
        capacity_status_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        process_unit: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch annual crude runs by refinery.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        quarter : Optional[Union[int, list[int], Series[int]]], optional
            filter by quarter, by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        owner : Optional[Union[str, list[str], Series[str]]], optional
            filter by Owner/Name, by default None
        capacity_status_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by CapacityStatusId, by default None
        process_unit : Optional[Union[str, list[str], Series[str]]], optional
            filter by ProcessUnit/Name, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Country/Name, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Region/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_runs(owner="BP")

        **Using Lists**
        >>> ci.WorldRefineryData().get_runs(year_gte=2020, refinery_id=[18, 357])

        **Using Series**
        >>> df = ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.ProcessUnits)
        >>> ci.WorldRefineryData().get_runs(year=2020, process_unit=df["Name"][:-10], refineryId=1)
        """
        endpoint = "runs"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("Year", year))
        filter_params.append(odata_list_to_filter("Quarter", quarter))
        filter_params.append(odata_list_to_filter("RefineryId", refinery_id))
        filter_params.append(odata_list_to_filter("Owner/Name", owner))
        filter_params.append(
            odata_list_to_filter("CapacityStatusId", capacity_status_id)
        )
        filter_params.append(odata_list_to_filter("ProcessUnit/Name", process_unit))
        filter_params.append(odata_list_to_filter("Refinery/Country/Name", country))
        filter_params.append(odata_list_to_filter("Refinery/Region/Name", region))

        if year_gt:
            filter_params.append(f"year gt {year_gt}")
        if year_gte:
            filter_params.append(f"year ge {year_gte}")
        if year_lt:
            filter_params.append(f"year lt {year_lt}")
        if year_lte:
            filter_params.append(f"year le {year_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
            "$expand": "*",
        }

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._to_df,
            paginate_fn=self._paginate,
        )

    def get_yields(
        self,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        quarter: Optional[Union[int, list[int], "Series[int]"]] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        owner: Optional[Union[str, list[str], "Series[str]"]] = None,
        capacity_status_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        process_unit: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch refinery yields by year.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        quarter : Optional[Union[int, list[int], Series[int]]], optional
            filter by quarter, by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        owner : Optional[Union[str, list[str], Series[str]]], optional
            filter by Owner/Name, by default None
        capacity_status_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by CapacityStatusId, by default None
        process_unit : Optional[Union[str, list[str], Series[str]]], optional
            filter by ProcessUnit/Name, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Country/Name, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Region/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_yields(refineryId=1)

        **Using Lists**
        >>> ci.WorldRefineryData().get_yields(year=2020, owner=["BP", "S N Repal"])

        **Using Series**
        >>> df = ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.ProcessUnits)
        >>> ci.WorldRefineryData().get_yields(year=2020, process_unit=df["Name"][:-10], refineryId=1)
        """
        endpoint = "yields"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("Year", year))
        filter_params.append(odata_list_to_filter("Quarter", quarter))
        filter_params.append(odata_list_to_filter("RefineryId", refinery_id))
        filter_params.append(odata_list_to_filter("Owner/Name", owner))
        filter_params.append(
            odata_list_to_filter("CapacityStatusId", capacity_status_id)
        )
        filter_params.append(odata_list_to_filter("ProcessUnit/Name", process_unit))
        filter_params.append(odata_list_to_filter("Refinery/Country/Name", country))
        filter_params.append(odata_list_to_filter("Refinery/Region/Name", region))

        if year_gt:
            filter_params.append(f"year gt {year_gt}")
        if year_gte:
            filter_params.append(f"year ge {year_gte}")
        if year_lt:
            filter_params.append(f"year lt {year_lt}")
        if year_lte:
            filter_params.append(f"year le {year_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
            "$expand": "*",
        }

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._to_df,
            paginate_fn=self._paginate,
        )

    @staticmethod
    def _outage_alert_pg_fn(resp: Response) -> Paginator:
        j = resp.json()

        count: int = j["metadata"]["count"]
        page_size: int = j["metadata"]["pageSize"]

        if not page_size:
            return Paginator(False, "page", 0)

        remainder = count % int(page_size)
        quotient = count // int(page_size)
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", total_pages)

        return Paginator(True, "page", total_pages)

    @staticmethod
    def _outage_alerts_to_df(resp: Response) -> DataFrame:
        j = resp.json()

        df = json_normalize(
            j["results"],
            meta=[
                "outageId",
                "refineryId",
                "operatorname",
                "countryName",
                "cityName",
                "latitude",
                "longitude",
                "createdDate",
            ],
            record_path="alerts",
        )
        df["createdDate"] = to_datetime(df["createdDate"], utc=True)
        df["modifiedDate"] = to_datetime(df["modifiedDate"], utc=True)
        df["startDate"] = to_datetime(df["startDate"])
        df["endDate"] = to_datetime(df["endDate"])
        return df

    def get_outage_alerts(
        self,
        *,
        outage_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        created_date: Optional[Union[date, list[date], "Series[date]"]] = None,
        created_date_gt: Optional[date] = None,
        created_date_gte: Optional[date] = None,
        created_date_lt: Optional[date] = None,
        created_date_lte: Optional[date] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        operator: Optional[Union[str, list[str], "Series[str]"]] = None,
        planning_status: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        process_unit: Optional[Union[str, list[str], "Series[str]"]] = None,
        outage_vol: Optional[Union[float, list[float], "Series[float]"]] = None,
        outage_vol_gt: Optional[float] = None,
        outage_vol_gte: Optional[float] = None,
        outage_vol_lt: Optional[float] = None,
        outage_vol_lte: Optional[float] = None,
        # modified_date: Optional[Union[date, list[date], "Series[date]"]] = None,
        # modified_date_gt: Optional[date] = None,
        # modified_date_gte: Optional[date] = None,
        # modified_date_lt: Optional[date] = None,
        # modified_date_lte: Optional[date] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch intra-day refinery outages.

        Parameters
        ----------
        outage_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by outageId, by default None
        created_date : Optional[Union[date, list[date], Series[date]]], optional
            _description_, by default None
        created_date_gt : Optional[date], optional
            _description_, by default None
        created_date_gte : Optional[date], optional
            _description_, by default None
        created_date_lt : Optional[date], optional
            _description_, by default None
        created_date_lte : Optional[date], optional
            _description_, by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        operator : Optional[Union[str, list[str], Series[str]]], optional
            filter by operatorname, by default None
        planning_status : Optional[Union[str, list[str], Series[str]]], optional
            filter by planningStatus, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by countryName, by default None
        process_unit : Optional[Union[str, list[str], Series[str]]], optional
            filter by processUnitName, by default None
        outage_vol : Optional[Union[float, list[float], Series[float]]], optional
            filter by ``OutageVol_MBD = x`` , by default None
        outage_vol_gt : Optional[float], optional
            filter by ``OutageVol_MBD > x`` , by default None
        outage_vol_gte : Optional[float], optional
            filter by ``OutageVol_MBD >= x`` , by default None
        outage_vol_lt : Optional[float], optional
            filter by ``OutageVol_MBD < x`` , by default None
        outage_vol_lte : Optional[float], optional
            filter by ``OutageVol_MBD <= x`` , by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to skip a certain number of records, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_outage_alerts(refineryId=245)

        **Using Lists**
        >>> ci.WorldRefineryData().get_outage_alerts(process_unit=["Coker", "CDU"])
        """
        endpoint = "outage-alerts"

        filter_params: list[str] = []

        filter_params.append(list_to_filter("outageId", outage_id))
        filter_params.append(list_to_filter("refineryId", refinery_id))
        filter_params.append(list_to_filter("operatorname", operator))
        filter_params.append(list_to_filter("countryName", country))
        filter_params.append(list_to_filter("planningStatus", planning_status))
        filter_params.append(list_to_filter("processUnitName", process_unit))
        filter_params.append(list_to_filter("outageVol_MBD", outage_vol))
        # filter_params.append(list_to_filter("modifiedDate", modified_date))
        filter_params.append(list_to_filter("createdDate", created_date))

        if created_date_gt:
            filter_params.append(f'createdDate > "{created_date_gt}"')
        if created_date_gte:
            filter_params.append(f'createdDate >= "{created_date_gte}"')
        if created_date_lt:
            filter_params.append(f'createdDate < "{created_date_lt}"')
        if created_date_lte:
            filter_params.append(f'createdDate <= "{created_date_lte}"')

        # if modified_date_gt:
        #     filter_params.append(f'modifiedDate > "{modified_date_gt}"')
        # if modified_date_gte:
        #     filter_params.append(f'modifiedDate >= "{modified_date_gte}"')
        # if modified_date_lt:
        #     filter_params.append(f'modifiedDate < "{modified_date_lt}"')
        # if modified_date_lte:
        #     filter_params.append(f'modifiedDate <= "{modified_date_lte}"')

        if outage_vol_gt:
            filter_params.append(f"OutageVol_MBD > {outage_vol_gt}")
        if outage_vol_gte:
            filter_params.append(f"OutageVol_MBD >= {outage_vol_gte}")
        if outage_vol_lt:
            filter_params.append(f"OutageVol_MBD < {outage_vol_lt}")
        if outage_vol_lte:
            filter_params.append(f"OutageVol_MBD <= {outage_vol_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params: dict[str, Union[str, int]] = {
            "page": page,
            "pageSize": page_size,
        }

        if filter_exp:
            params["filter"] = filter_exp

        return get_data(
            path=f"refinery-data/v1.2/{endpoint}",
            params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._outage_alerts_to_df,
            paginate_fn=self._outage_alert_pg_fn,
        )

    def get_outages(
        self,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        date: Optional[Union[date, list[date], "Series[date]"]] = None,
        date_gt: Optional[date] = None,
        date_gte: Optional[date] = None,
        date_lt: Optional[date] = None,
        date_lte: Optional[date] = None,
        quarter: Optional[Union[int, list[int], "Series[int]"]] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        owner: Optional[Union[str, list[str], "Series[str]"]] = None,
        process_unit: Optional[Union[str, list[str], "Series[str]"]] = None,
        planning_status: Optional[Union[str, list[str], "Series[str]"]] = None,
        outage_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        outage_vol: Optional[Union[float, list[float], "Series[float]"]] = None,
        outage_vol_gt: Optional[float] = None,
        outage_vol_gte: Optional[float] = None,
        outage_vol_lt: Optional[float] = None,
        outage_vol_lte: Optional[float] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch refinery outages.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x`` , by default None
        year_gt : Optional[int], optional
            filter by ``year > x`` , by default None
        year_gte : Optional[int], optional
            filter by ``year >= x`` , by default None
        year_lt : Optional[int], optional
            filter by ``year < x`` , by default None
        year_lte : Optional[int], optional
            filter by ``year <= x`` , by default None
        date : Optional[Union[date, list[date], Series[date]]], optional
            filter by ``date = x`` , by default None
        date_gt : Optional[date], optional
            filter by ``date > x`` , by default None
        date_gte : Optional[date], optional
            filter by ``date >= x`` , by default None
        date_lt : Optional[date], optional
            filter by ``date < x`` , by default None
        date_lte : Optional[date], optional
            filter by ``date <= x`` , by default None
        quarter : Optional[Union[int, list[int], Series[int]]], optional
            filter by quarter, by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        owner : Optional[Union[str, list[str], Series[str]]], optional
            filter by Owner/Name, by default None
        process_unit : Optional[Union[str, list[str], Series[str]]], optional
            filter by ProcessUnit/Name, by default None
        planning_status : Optional[Union[str, list[str], Series[str]]], optional
            filter by PlanningStatus, by default None
        outage_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by outageId, by default None
        outage_vol : Optional[Union[float, list[float], Series[float]]], optional
            filter by ``OutageVol_MBD = x`` , by default None
        outage_vol_gt : Optional[float], optional
            filter by ``OutageVol_MBD > x`` , by default None
        outage_vol_gte : Optional[float], optional
            filter by ``OutageVol_MBD >= x`` , by default None
        outage_vol_lt : Optional[float], optional
            filter by ``OutageVol_MBD < x`` , by default None
        outage_vol_lte : Optional[float], optional
            filter by ``OutageVol_MBD <= x`` , by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Country/Name, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Region/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_outages(year=2020, quarter=3)

        **Using Lists**
        >>> ci.WorldRefineryData().get_outages(year=2020, planning_status="Unplanned", process_unit=["Coker", "CDU"])

        **Using Series**
        >>> df = ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.Refineries)
        >>> df = df[df['Country.Name'] == "Mexico"]
        >>> ci.WorldRefineryData().get_outages(outage_vol_gte=200, refinery_id=df['Id'])
        """
        endpoint = "outages"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("Year", year))
        filter_params.append(odata_list_to_filter("Date", date))
        filter_params.append(odata_list_to_filter("OutageVol_MBD", outage_vol))
        filter_params.append(odata_list_to_filter("Quarter", quarter))
        filter_params.append(odata_list_to_filter("RefineryId", refinery_id))
        filter_params.append(odata_list_to_filter("Owner/Name", owner))
        filter_params.append(odata_list_to_filter("ProcessUnit/Name", process_unit))
        filter_params.append(odata_list_to_filter("PlanningStatus", planning_status))
        filter_params.append(odata_list_to_filter("OutageId", outage_id))
        filter_params.append(odata_list_to_filter("Refinery/Country/Name", country))
        filter_params.append(odata_list_to_filter("Refinery/Region/Name", region))

        if year_gt:
            filter_params.append(f"year gt {year_gt}")
        if year_gte:
            filter_params.append(f"year ge {year_gte}")
        if year_lt:
            filter_params.append(f"year lt {year_lt}")
        if year_lte:
            filter_params.append(f"year le {year_lte}")

        if date_gt:
            filter_params.append(f"Date gt {date_gt}")
        if date_gte:
            filter_params.append(f"Date ge {date_gte}")
        if date_lt:
            filter_params.append(f"Date lt {date_lt}")
        if date_lte:
            filter_params.append(f"Date le {date_lte}")

        if outage_vol_gt:
            filter_params.append(f"OutageVol_MBD gt {outage_vol_gt}")
        if outage_vol_gte:
            filter_params.append(f"OutageVol_MBD ge {outage_vol_gte}")
        if outage_vol_lt:
            filter_params.append(f"OutageVol_MBD lt {outage_vol_lt}")
        if outage_vol_lte:
            filter_params.append(f"OutageVol_MBD le {outage_vol_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
            "$expand": "*",
        }

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._to_df,
            paginate_fn=self._paginate,
        )

    def get_ownership(
        self,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        quarter: Optional[Union[int, list[int], "Series[int]"]] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        owner: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch refinery ownership

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x`` , by default None
        year_gt : Optional[int], optional
            filter by ``year > x`` , by default None
        year_gte : Optional[int], optional
            filter by ``year >= x`` , by default None
        year_lt : Optional[int], optional
            filter by ``year < x`` , by default None
        year_lte : Optional[int], optional
            filter by ``year <= x`` , by default None
        quarter : Optional[Union[int, list[int], Series[int]]], optional
            filter by quarter, by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        owner : Optional[Union[str, list[str], Series[str]]], optional
            filter by Owner/Name, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Country/Name, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Region/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_ownership(owner="BP")

        **Using Lists**
        >>> ci.WorldRefineryData().get_ownership(year_gte=2017, quarter=[3, 4])

        **Using Series**
        >>> df = ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.Owners)
        >>> df = df[df['Name'].str.contains("Shell")]
        >>> ci.WorldRefineryData().get_ownership(owner=df['Name'])
        """
        endpoint = "ownership"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("Year", year))
        filter_params.append(odata_list_to_filter("Quarter", quarter))
        filter_params.append(odata_list_to_filter("RefineryId", refinery_id))
        filter_params.append(odata_list_to_filter("Owner/Name", owner))
        filter_params.append(odata_list_to_filter("Refinery/Country/Name", country))
        filter_params.append(odata_list_to_filter("Refinery/Region/Name", region))

        if year_gt:
            filter_params.append(f"year gt {year_gt}")
        if year_gte:
            filter_params.append(f"year ge {year_gte}")
        if year_lt:
            filter_params.append(f"year lt {year_lt}")
        if year_lte:
            filter_params.append(f"year le {year_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
            "$expand": "*",
        }

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._to_df,
            paginate_fn=self._paginate,
        )

    def get_margins(
        self,
        *,
        date: Optional[Union[date, list[date], "Series[date]"]] = None,
        date_gt: Optional[date] = None,
        date_gte: Optional[date] = None,
        date_lt: Optional[date] = None,
        date_lte: Optional[date] = None,
        margin_type: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Fetch refinery margins

        Parameters
        ----------
        date : Optional[Union[date, list[date], Series[date]]], optional
            filter by ``date = x`` , by default None
        date_gt : Optional[date], optional
            filter by ``date > x`` , by default None
        date_gte : Optional[date], optional
            filter by ``date >= x`` , by default None
        date_lt : Optional[date], optional
            filter by ``date < x`` , by default None
        date_lte : Optional[date], optional
            filter by ``date <= x`` , by default None
        margin_type : Optional[Union[str, list[str], Series[str]]], optional
            filter by MarginType/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_margins(date=date(2023, 2, 17))

        **Using Lists**
        >>> ci.WorldRefineryData().get_margins(date_gte=date(2023, 1, 1), margin_type=['Dated Brent NWE Cracking', 'Dubai Singapore Cracking'])

        **Using Series**
        >>> df = ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.MarginTypes)
        >>> df = df[df['Name'].str.contains("Brent")]
        >>> ci.WorldRefineryData().get_ownership(date_gte=date(2023, 1, 1), margin_type=df['Name'])
        """
        endpoint = "margins"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("Date", date))
        filter_params.append(odata_list_to_filter("MarginType/Name", margin_type))

        if date_gt:
            filter_params.append(f"Date gt {date_gt}")
        if date_gte:
            filter_params.append(f"Date ge {date_gte}")
        if date_lt:
            filter_params.append(f"Date lt {date_lt}")
        if date_lte:
            filter_params.append(f"Date le {date_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
            "$expand": "*",
        }

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._to_df,
            paginate_fn=self._paginate,
        )

    def get_reference_data(
        self, type: RefTypes, raw: bool = False
    ) -> Union[Response, DataFrame]:
        """
        Fetch reference data for the World Refinery Database.

        Parameters
        ----------
        type : RefTypes
            filter by type
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_reference_data(type=ci.WorldRefineryData.RefTypes.Owners)
        """
        endpoint_path = type.value

        params = {"pageSize": 1000, "$skip": 0, "$expand": "*", "$count": "true"}

        if type == self.RefTypes.PADDs:
            df_fn = self._padd_to_df
        else:
            df_fn = self._to_df

        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint_path}?{qs}",
            params={},
            # params=params,
            paginate=True,
            raw=raw,
            paginate_fn=self._paginate,
            df_fn=df_fn,
        )

    def get_outages_grouped(
        self,
        *,
        start_date: Optional[Union[date, list[date], "Series[date]"]] = None,
        start_date_gt: Optional[date] = None,
        start_date_gte: Optional[date] = None,
        start_date_lt: Optional[date] = None,
        start_date_lte: Optional[date] = None,
        end_date: Optional[Union[date, list[date], "Series[date]"]] = None,
        end_date_gt: Optional[date] = None,
        end_date_gte: Optional[date] = None,
        end_date_lt: Optional[date] = None,
        end_date_lte: Optional[date] = None,
        refinery_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        owner: Optional[Union[str, list[str], "Series[str]"]] = None,
        process_unit: Optional[Union[str, list[str], "Series[str]"]] = None,
        planning_status: Optional[Union[str, list[str], "Series[str]"]] = None,
        outage_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        outage_vol: Optional[Union[float, list[float], "Series[float]"]] = None,
        outage_vol_gt: Optional[float] = None,
        outage_vol_gte: Optional[float] = None,
        outage_vol_lt: Optional[float] = None,
        outage_vol_lte: Optional[float] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        skip: int = 0,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Transformation of refinery outages endpoint so that each outage-owner is a single row with outage `start_date` and `end_date` as columns

        Parameters
        ----------
        start_date : Optional[Union[date, list[date], Series[date]]], optional
            filter by ``start_date = x`` , by default None
        start_date_gt : Optional[date], optional
            filter by ``start_date > x`` , by default None
        start_date_gte : Optional[date], optional
            filter by ``start_date >= x`` , by default None
        start_date_lt : Optional[date], optional
            filter by ``start_date < x`` , by default None
        start_date_lte : Optional[date], optional
            filter by ``start_date <= x`` , by default None
        end_date : Optional[Union[date, list[date], Series[date]]], optional
            filter by ``end_date = x`` , by default None
        end_date_gt : Optional[date], optional
            filter by ``end_date > x`` , by default None
        end_date_gte : Optional[date], optional
            filter by ``end_date >= x`` , by default None
        end_date_lt : Optional[date], optional
            filter by ``end_date < x`` , by default None
        end_date_lte : Optional[date], optional
            filter by ``end_date <= x`` , by default None
        refinery_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by refineryId, by default None
        owner : Optional[Union[str, list[str], Series[str]]], optional
            filter by Owner/Name, by default None
        process_unit : Optional[Union[str, list[str], Series[str]]], optional
            filter by ProcessUnit/Name, by default None
        planning_status : Optional[Union[str, list[str], Series[str]]], optional
            filter by PlanningStatus, by default None
        outage_id : Optional[Union[int, list[int], Series[int]]], optional
            filter by outageId, by default None
        outage_vol : Optional[Union[float, list[float], Series[float]]], optional
            filter by ``OutageVol_MBD = x`` , by default None
        outage_vol_gt : Optional[float], optional
            filter by ``OutageVol_MBD > x`` , by default None
        outage_vol_gte : Optional[float], optional
            filter by ``OutageVol_MBD >= x`` , by default None
        outage_vol_lt : Optional[float], optional
            filter by ``OutageVol_MBD < x`` , by default None
        outage_vol_lte : Optional[float], optional
            filter by ``OutageVol_MBD <= x`` , by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Country/Name, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by Refinery/Region/Name, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        skip : int, optional
            pass-thru ``$skip`` query param to skip a certain number of records, by default 0
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.WorldRefineryData().get_outages_grouped(country="United States")

        **Using Lists**
        >>> ci.WorldRefineryData().get_outages_grouped(process_unit=["Coker", "CDU"])

        """
        endpoint = "outages"

        filter_params: list[str] = []

        filter_params.append(odata_list_to_filter("start_date", start_date))
        filter_params.append(odata_list_to_filter("end_date", end_date))
        filter_params.append(odata_list_to_filter("OutageVol_MBD", outage_vol))
        filter_params.append(odata_list_to_filter("RefineryId", refinery_id))
        filter_params.append(odata_list_to_filter("Owner/Name", owner))
        filter_params.append(odata_list_to_filter("ProcessUnit/Name", process_unit))
        filter_params.append(odata_list_to_filter("PlanningStatus", planning_status))
        filter_params.append(odata_list_to_filter("OutageId", outage_id))
        filter_params.append(odata_list_to_filter("Refinery/Country/Name", country))
        filter_params.append(odata_list_to_filter("Refinery/Region/Name", region))

        if start_date_gt:
            filter_params.append(f"start_date gt {start_date_gt}")
        if start_date_gte:
            filter_params.append(f"start_date ge {start_date_gte}")
        if start_date_lt:
            filter_params.append(f"start_date lt {start_date_lt}")
        if start_date_lte:
            filter_params.append(f"start_date le {start_date_lte}")

        if end_date_gt:
            filter_params.append(f"end_date gt {end_date_gt}")
        if end_date_gte:
            filter_params.append(f"end_date ge {end_date_gte}")
        if end_date_lt:
            filter_params.append(f"end_date lt {end_date_lt}")
        if end_date_lte:
            filter_params.append(f"end_date le {end_date_lte}")

        if outage_vol_gt:
            filter_params.append(f"OutageVol_MBD gt {outage_vol_gt}")
        if outage_vol_gte:
            filter_params.append(f"OutageVol_MBD ge {outage_vol_gte}")
        if outage_vol_lt:
            filter_params.append(f"OutageVol_MBD lt {outage_vol_lt}")
        if outage_vol_lte:
            filter_params.append(f"OutageVol_MBD le {outage_vol_lte}")

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "$skip": skip,
            "pageSize": page_size,
            "$count": "true",
        }

        groupby = "OutageId, RefineryId, Owner/Name, Refinery/Country/Name, Refinery/Region/Name, PlanningStatus, ProcessUnit/Name"
        aggregate = "OutageVol_MBD with max as OutageVol_MBD, Date with max as end_date, Date with min as start_date"
        apply_string = f"groupby(({groupby}), aggregate({aggregate}))"
        params["$apply"] = apply_string

        if filter_exp:
            params["$filter"] = filter_exp

        # odata endpoint will not allow '+' character in URL seemingly
        qs = urlencode(params, quote_via=quote)

        return get_data(
            path=f"{self._endpoint}{endpoint}?{qs}",
            params={},
            # params=params,
            paginate=paginate,
            raw=raw,
            df_fn=self._ts_to_df,
            paginate_fn=self._paginate,
        )



# Copyright 2023 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from spgci.api_client import get_data, Paginator
from typing import List, Union, Optional
from requests import Response
from spgci.utilities import list_to_filter
from pandas import Series, DataFrame, to_datetime  # type: ignore
from datetime import date, datetime
from enum import Enum


class EWindowMarketData:
    """
    EWindow Market Data - Bids, Offers, Trades

    Includes
    --------
    ``OrderState`` enum for order states.\n
    ``OrderType`` enum for order types.\n
    ``get_products()`` to see the list of Products and their latest order date.\n
    ``get_markets()`` to see the list of Markets and their latest order date.\n
    ``get_botes()`` to get Bids, Offers, Trades.\n

    """

    _path = "tradedata/v3/"

    class OrderState(Enum):
        """Order State"""

        Active = "active"
        Consummated = "consummated"
        Inactive = "inactive"
        Withdrawn = "withdrawn"

    class OrderType(Enum):
        """Order Type"""

        Bid = "bid"
        Offer = "offer"
        RFBid = "rf bid"
        RFOffer = "rf offer"

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        total_pages = j["metadata"]["total_pages"]

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages)

    @staticmethod
    def _convert_to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = DataFrame(j["results"])

        if len(df) > 0:
            df["order_begin"] = to_datetime(df["order_begin"])
            df["order_end"] = to_datetime(df["order_end"])
            df["order_date"] = to_datetime(df["order_date"])
            df["order_time"] = to_datetime(df["order_time"])
            df["deal_begin"] = to_datetime(df["deal_begin"])
            df["deal_end"] = to_datetime(df["deal_end"])

        return df

    @staticmethod
    def _convert_agg_to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = DataFrame(j["aggResultValue"])

        if len(df) > 0:
            df["max(order_date)"] = to_datetime(df["max(order_date)"])

        return df

    def get_markets(self, raw: bool = False) -> Union[DataFrame, Response]:
        """
        Fetch the list of Markets.

        Parameters
        ----------
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.EWindowMarketData().get_markets()
        """
        params = {
            "groupBy": "market",
            "field": "max(order_date)",
            "pageSize": 1000,
            "sort": "order_time:desc",
        }

        return get_data(
            path=f"{self._path}ewindowdata",
            raw=raw,
            params=params,
            paginate=True,
            paginate_fn=self._paginate,
            df_fn=self._convert_agg_to_df,
        )

    def get_products(self, raw: bool = False) -> Union[DataFrame, Response]:
        """
        Fetch the list of Products.

        Parameters
        ----------
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.EWindowMarketData().get_products()
        """
        params = {
            "groupBy": "product, market",
            "field": "max(order_date)",
            "pageSize": 1000,
            "sort": "order_time:desc",
        }

        return get_data(
            path=f"{self._path}ewindowdata",
            raw=raw,
            params=params,
            paginate=True,
            paginate_fn=self._paginate,
            df_fn=self._convert_agg_to_df,
        )

    def get_botes(
        self,
        *,
        market: Optional[Union[list[str], "Series[str]", str]] = None,
        product: Optional[Union[list[str], "Series[str]", str]] = None,
        hub: Optional[Union[list[str], "Series[str]", str]] = None,
        strip: Optional[Union[list[str], "Series[str]", str]] = None,
        buyer: Optional[Union[list[str], "Series[str]", str]] = None,
        seller: Optional[Union[list[str], "Series[str]", str]] = None,
        order_type: Optional[
            Union[list[str], list[OrderType], "Series[str]", str, OrderType]
        ] = None,
        order_state: Optional[
            Union[list[str], list[OrderState], "Series[str]", str, OrderState]
        ] = None,
        order_id: Optional[Union[list[int], "Series[int]", int]] = None,
        order_time: Optional[Union[datetime, date]] = None,
        order_time_lt: Optional[Union[datetime, date]] = None,
        order_time_lte: Optional[Union[datetime, date]] = None,
        order_time_gt: Optional[Union[datetime, date]] = None,
        order_time_gte: Optional[Union[datetime, date]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Fetch BOTes (Bids, Offers, Trades) from the EWindow MarketData API.

        See ``get_products()`` to search for products.\n
        See ``get_markets()`` to search for markets.\n

        Parameters
        ----------
        market : Optional[Union[list[str], Series[str], str]], optional
            filter by market, by default None
        product : Optional[Union[list[str], Series[str], str]], optional
            filter by product, by default None
        hub : Optional[Union[list[str], Series[str], str]], optional
            filter by hub, by default None
        strip : Optional[Union[list[str], Series[str], str]], optional
            filter by strip, by default None
        buyer : Optional[Union[list[str], Series[str], str]], optional
            filter by buyer, by default None
        seller : Optional[Union[list[str], Series[str], str]], optional
            filter by seller, by default None
        order_type : Optional[Union[list[str], list[OrderType], Series[str], str, OrderType]], optional
            filter by order type, by default None
        order_state : Optional[ Union[list[str], list[OrderState], Series[str], str, OrderState] ], optional
            filter by order state, by default None
        order_id : Optional[Union[list[int], Series[int], int]], optional
            filter by order id, by default None
        order_time : Optional[Union[datetime, date]], optional
            filter by ``order_time = x``, by default None
        order_time_lt : Optional[Union[datetime, date]], optional
            filter by ``order_time < x``, by default None
        order_time_lte : Optional[Union[datetime, date]], optional
            filter by ``order_time <= x``, by default None
        order_time_gt : Optional[Union[datetime, date]], optional
            filter by ``order_time > x``, by default None
        order_time_gte : Optional[Union[datetime, date]], optional
            filter by ``order_time >= x``, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> today = date.today()
        >>> ci.EWindowMarketData().get_botes(market=["EU BFOE", "US MidWest"], order_time_gte=today)

        **Date Range**
        >>> d1 = date(2023, 2, 1)
        >>> d2 = date(2023, 2, 3)
        >>> ci.EWindowMarketData().get_botes(market=["EU BFOE", "US MidWest"], order_time_gt=d1, order_time_lt=d2)
        """
        endpoint_path = "ewindowdata"

        filter_params: List[str] = []

        filter_params.append(list_to_filter("market", market))
        filter_params.append(list_to_filter("product", product))
        filter_params.append(list_to_filter("hub", hub))
        filter_params.append(list_to_filter("strip", strip))
        filter_params.append(list_to_filter("buyer", buyer))
        filter_params.append(list_to_filter("seller", seller))
        filter_params.append(list_to_filter("order_type", order_type))
        filter_params.append(list_to_filter("order_id", order_id))
        filter_params.append(list_to_filter("order_state", order_state))

        if order_time:
            filter_params.append(f'order_time: "{order_time}"')
        if order_time_gt:
            filter_params.append(f'order_time > "{order_time_gt}"')
        if order_time_gte:
            filter_params.append(f'order_time >= "{order_time_gte}"')
        if order_time_lt:
            filter_params.append(f'order_time < "{order_time_lt}"')
        if order_time_lte:
            filter_params.append(f'order_time <= "{order_time_lte}"')

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            raw=raw,
            df_fn=self._convert_to_df,
            paginate_fn=self._paginate,
            paginate=paginate,
        )

        return response



from __future__ import annotations
from typing import List, Optional, Union
from requests import Response
from spgci.api_client import get_data
from spgci.utilities import list_to_filter
from pandas import DataFrame, Series
from datetime import date, datetime
import pandas as pd
from dateutil.parser import parse
from typing import Literal

class AmericasGas:

    _datasets = Literal[
    "pipeline-flows",
    ]

    def get_reference_data_pipeline_flows(
        self,
        *,
        last_modified_date: Optional[datetime] = None,
        last_modified_date_lt: Optional[datetime] = None,
        last_modified_date_lte: Optional[datetime] = None,
        last_modified_date_gt: Optional[datetime] = None,
        last_modified_date_gte: Optional[datetime] = None,
        legacy_point_logic_point_id: Optional[
            Union[list[str], Series[str], str]
        ] = None,
        legacy_point_logic_lci_id: Optional[Union[list[str], Series[str], str]] = None,
        legacy_bentek_point_id: Optional[Union[list[str], Series[str], str]] = None,
        component_point_id: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_operator_id: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_operator_name: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_id: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_name: Optional[Union[list[str], Series[str], str]] = None,
        point_name: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_primary: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_id_primary: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_id_secondary: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_secondary: Optional[Union[list[str], Series[str], str]] = None,
        flow_direction: Optional[Union[list[str], Series[str], str]] = None,
        flow_direction_code: Optional[Union[list[str], Series[str], str]] = None,
        flow_direction_id: Optional[Union[list[str], Series[str], str]] = None,
        zone: Optional[Union[list[str], Series[str], str]] = None,
        company_id: Optional[Union[list[str], Series[str], str]] = None,
        connecting_party: Optional[Union[list[str], Series[str], str]] = None,
        loc_prop: Optional[Union[list[str], Series[str], str]] = None,
        point_is_active: Optional[Union[list[str], Series[str], str]] = None,
        domain: Optional[Union[list[str], Series[str], str]] = None,
        domain_id: Optional[Union[list[str], Series[str], str]] = None,
        region_id: Optional[Union[list[str], Series[str], str]] = None,
        region: Optional[Union[list[str], Series[str], str]] = None,
        subregion_id: Optional[Union[list[str], Series[str], str]] = None,
        subregion: Optional[Union[list[str], Series[str], str]] = None,
        state_id: Optional[Union[list[str], Series[str], str]] = None,
        state: Optional[Union[list[str], Series[str], str]] = None,
        county_id: Optional[Union[list[str], Series[str], str]] = None,
        county: Optional[Union[list[str], Series[str], str]] = None,
        producing_area: Optional[Union[list[str], Series[str], str]] = None,
        producing_area_id: Optional[Union[list[str], Series[str], str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 5000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        A single source for pipeline metadata that captures 35 different attributes. Mimics pipeline flows dataset with the removal of volume metric time series data.

        Parameters
        ----------

         last_modified_date: Optional[datetime], optional
             Date and time the record was last updated., by default None
         last_modified_date_gt: Optional[datetime], optional
             filter by `last_modified_date > x`, by default None
         last_modified_date_gte: Optional[datetime], optional
             filter by `last_modified_date >= x`, by default None
         last_modified_date_lt: Optional[datetime], optional
             filter by `last_modified_date < x`, by default None
         last_modified_date_lte: Optional[datetime], optional
             filter by `last_modified_date <= x`, by default None
         legacy_point_logic_point_id: Optional[Union[list[str], Series[str], str]]
             Point ID for a meter used by the legacy PointLogic service., by default None
         legacy_point_logic_lci_id: Optional[Union[list[str], Series[str], str]]
             Alternative point ID for a meter used by the legacy PointLogic service., by default None
         legacy_bentek_point_id: Optional[Union[list[str], Series[str], str]]
             Point ID for a meter used by the legacy Bentek service., by default None
         component_point_id: Optional[Union[list[str], Series[str], str]]
             Point ID for a meter used by the Americas Gas service., by default None
         pipeline_operator_id: Optional[Union[list[str], Series[str], str]]
             ID associated with the common parent owner or operator of a pipeline system., by default None
         pipeline_operator_name: Optional[Union[list[str], Series[str], str]]
             The name of the common parent owner or operator of a pipeline system., by default None
         pipeline_id: Optional[Union[list[str], Series[str], str]]
             The ID given to a pipeline system, utilizes legacy Bentek pipeline ids when applicable., by default None
         pipeline_name: Optional[Union[list[str], Series[str], str]]
             The display name of a pipeline system, utilizes legacy Bentek names when applicable., by default None
         point_name: Optional[Union[list[str], Series[str], str]]
             The display name of a meter or point, utilizes legacy Bentek point name when applicable., by default None
         meter_type_primary: Optional[Union[list[str], Series[str], str]]
             The primary type of classification and purpose of a meter or point, utilizes legacy PointLogic definitions., by default None
         meter_type_id_primary: Optional[Union[list[str], Series[str], str]]
             An ID for the primary type of classification and purpose of a meter or point, utilizes legacy PointLogic definitions and ID., by default None
         meter_type_id_secondary: Optional[Union[list[str], Series[str], str]]
             A secondary type classification and purpose of a meter or point, meant to provide an extra level of detail and utilizes legacy Bentek ids., by default None
         meter_type_secondary: Optional[Union[list[str], Series[str], str]]
             A secondary type classification and purpose of a meter or point, meant to provide an extra level of detail, utilizes legacy Bentek definitions., by default None
         flow_direction: Optional[Union[list[str], Series[str], str]]
             Flow direction indicates the orientation of a point such as receipt, delivery, bi-directional or the reported flow direction of segment or compressor. Attribute is sourced from legacy PointLogic. Flow direction is primary to the similar and secondary attribute of Location Type., by default None
         flow_direction_code: Optional[Union[list[str], Series[str], str]]
             A one letter code for Flow Direction such as ‘R’ for receipt or ‘D’ for delivery. Attribute is sourced from legacy PointLogic. Flow direction is primary to the similar and secondary attribute of Location Type., by default None
         flow_direction_id: Optional[Union[list[str], Series[str], str]]
             Flow direction identification number for the orientation of a point such as receipt, delivery, bi-directional or the reported flow direction of segment or compressor. Attribute is sourced from legacy PointLogic. Flow direction is primary to the similar and secondary attribute of Location Type., by default None
         zone: Optional[Union[list[str], Series[str], str]]
             A designation for where on the pipeline system the point is located, corresponding to a pipeline’s operational and market design. Zonal information is sourced from the legacy Bentek service., by default None
         company_id: Optional[Union[list[str], Series[str], str]]
             An ID sourced from the legacy Bentek service used to identify the connecting business or company name of a meter., by default None
         connecting_party: Optional[Union[list[str], Series[str], str]]
             The downstream connecting business name of a meter as reported by the pipeline, utilizes legacy Bentek service., by default None
         loc_prop: Optional[Union[list[str], Series[str], str]]
             The location propriety code reported by the pipeline for a specific meter., by default None
         point_is_active: Optional[Union[list[str], Series[str], str]]
             A true or false return if a point or meter is active and in use within the Americas Gas service., by default None
         domain: Optional[Union[list[str], Series[str], str]]
             US Lower-48, Canada and Mexico., by default None
         domain_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the domain., by default None
         region_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the geographic region., by default None
         region: Optional[Union[list[str], Series[str], str]]
             A defined geographic region within the Americas Gas service. Regions are an aggregation of states or provinces within a country., by default None
         subregion_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the geographic subregion., by default None
         subregion: Optional[Union[list[str], Series[str], str]]
             A defined geographic subregion within the Americas Gas service. A substate geography is sometimes referred to as a subregion. Subregions are an aggregation of specific counties within a region and a country., by default None
         state_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the state or province utilizes legacy Bentek IDs., by default None
         state: Optional[Union[list[str], Series[str], str]]
             The political boundaries that define a state or province within country., by default None
         county_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the county within a US state in which a meter or point resides, utilizes legacy Bentek IDs., by default None
         county: Optional[Union[list[str], Series[str], str]]
             The political boundaries of a defined county within a US state in which a meter or point resides., by default None
         producing_area: Optional[Union[list[str], Series[str], str]]
             Defined aggregation of counties within a state that is a best fit representation of prominent oil and gas plays and basins., by default None
         producing_area_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for a defined Producing Area utilizes legacy PointLogic IDs., by default None
         filter_exp: Optional[str] = None,
         page: int = 1,
         page_size: int = 5000,
         raw: bool = False,
         paginate: bool = False

        """

        filter_params: List[str] = []
        filter_params.append(list_to_filter("lastModifiedDate", last_modified_date))
        if last_modified_date_gt is not None:
            filter_params.append(f'lastModifiedDate > "{last_modified_date_gt}"')
        if last_modified_date_gte is not None:
            filter_params.append(f'lastModifiedDate >= "{last_modified_date_gte}"')
        if last_modified_date_lt is not None:
            filter_params.append(f'lastModifiedDate < "{last_modified_date_lt}"')
        if last_modified_date_lte is not None:
            filter_params.append(f'lastModifiedDate <= "{last_modified_date_lte}"')
        filter_params.append(
            list_to_filter("legacyPointLogicPointId", legacy_point_logic_point_id)
        )
        filter_params.append(
            list_to_filter("legacyPointLogicLciId", legacy_point_logic_lci_id)
        )
        filter_params.append(
            list_to_filter("legacyBentekPointId", legacy_bentek_point_id)
        )
        filter_params.append(list_to_filter("componentPointId", component_point_id))
        filter_params.append(list_to_filter("pipelineOperatorId", pipeline_operator_id))
        filter_params.append(
            list_to_filter("pipelineOperatorName", pipeline_operator_name)
        )
        filter_params.append(list_to_filter("pipelineId", pipeline_id))
        filter_params.append(list_to_filter("pipelineName", pipeline_name))
        filter_params.append(list_to_filter("pointName", point_name))
        filter_params.append(list_to_filter("meterTypePrimary", meter_type_primary))
        filter_params.append(
            list_to_filter("meterTypeIdPrimary", meter_type_id_primary)
        )
        filter_params.append(
            list_to_filter("meterTypeIdSecondary", meter_type_id_secondary)
        )
        filter_params.append(list_to_filter("meterTypeSecondary", meter_type_secondary))
        filter_params.append(list_to_filter("flowDirection", flow_direction))
        filter_params.append(list_to_filter("flowDirectionCode", flow_direction_code))
        filter_params.append(list_to_filter("flowDirectionId", flow_direction_id))
        filter_params.append(list_to_filter("zone", zone))
        filter_params.append(list_to_filter("companyId", company_id))
        filter_params.append(list_to_filter("connectingParty", connecting_party))
        filter_params.append(list_to_filter("locProp", loc_prop))
        filter_params.append(list_to_filter("pointIsActive", point_is_active))
        filter_params.append(list_to_filter("domain", domain))
        filter_params.append(list_to_filter("domainId", domain_id))
        filter_params.append(list_to_filter("regionId", region_id))
        filter_params.append(list_to_filter("region", region))
        filter_params.append(list_to_filter("subregionId", subregion_id))
        filter_params.append(list_to_filter("subregion", subregion))
        filter_params.append(list_to_filter("stateId", state_id))
        filter_params.append(list_to_filter("state", state))
        filter_params.append(list_to_filter("countyId", county_id))
        filter_params.append(list_to_filter("county", county))
        filter_params.append(list_to_filter("producingArea", producing_area))
        filter_params.append(list_to_filter("producingAreaId", producing_area_id))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path=f"/analytics/gas/na-gas/v1/reference-data/pipeline-flows",
            params=params,
            df_fn=self._convert_to_df,
            raw=raw,
            paginate=paginate,
        )
        return response

    def get_pipeline_flows(
        self,
        *,
        flow_date: Optional[date] = None,
        flow_date_lt: Optional[date] = None,
        flow_date_lte: Optional[date] = None,
        flow_date_gt: Optional[date] = None,
        flow_date_gte: Optional[date] = None,
        last_modified_date: Optional[datetime] = None,
        last_modified_date_lt: Optional[datetime] = None,
        last_modified_date_lte: Optional[datetime] = None,
        last_modified_date_gt: Optional[datetime] = None,
        last_modified_date_gte: Optional[datetime] = None,
        date_frequency: Optional[Union[list[str], Series[str], str]] = None,
        date_frequency_desc: Optional[Union[list[str], Series[str], str]] = None,
        nomination_cycle: Optional[Union[list[str], Series[str], str]] = None,
        legacy_point_logic_point_id: Optional[
            Union[list[str], Series[str], str]
        ] = None,
        legacy_point_logic_lci_id: Optional[Union[list[str], Series[str], str]] = None,
        legacy_bentek_point_id: Optional[Union[list[str], Series[str], str]] = None,
        component_point_id: Optional[Union[list[str], Series[str], str]] = None,
        component_create_date: Optional[datetime] = None,
        component_create_date_lt: Optional[datetime] = None,
        component_create_date_lte: Optional[datetime] = None,
        component_create_date_gt: Optional[datetime] = None,
        component_create_date_gte: Optional[datetime] = None,
        pipeline_operator_id: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_operator_name: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_id: Optional[Union[list[str], Series[str], str]] = None,
        pipeline_name: Optional[Union[list[str], Series[str], str]] = None,
        point_name: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_primary: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_id_primary: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_id_secondary: Optional[Union[list[str], Series[str], str]] = None,
        meter_type_secondary: Optional[Union[list[str], Series[str], str]] = None,
        location_type_code: Optional[Union[list[str], Series[str], str]] = None,
        location_description: Optional[Union[list[str], Series[str], str]] = None,
        location_type_id: Optional[Union[list[str], Series[str], str]] = None,
        flow_direction: Optional[Union[list[str], Series[str], str]] = None,
        flow_direction_code: Optional[Union[list[str], Series[str], str]] = None,
        flow_direction_id: Optional[Union[list[str], Series[str], str]] = None,
        zone: Optional[Union[list[str], Series[str], str]] = None,
        company_id: Optional[Union[list[str], Series[str], str]] = None,
        connecting_party: Optional[Union[list[str], Series[str], str]] = None,
        loc_prop: Optional[Union[list[str], Series[str], str]] = None,
        point_is_active: Optional[Union[list[bool], Series[bool], bool]] = None,
        domain: Optional[Union[list[str], Series[str], str]] = None,
        domain_id: Optional[Union[list[str], Series[str], str]] = None,
        region_id: Optional[Union[list[str], Series[str], str]] = None,
        region: Optional[Union[list[str], Series[str], str]] = None,
        subregion_id: Optional[Union[list[str], Series[str], str]] = None,
        subregion: Optional[Union[list[str], Series[str], str]] = None,
        state_id: Optional[Union[list[str], Series[str], str]] = None,
        state: Optional[Union[list[str], Series[str], str]] = None,
        county_id: Optional[Union[list[str], Series[str], str]] = None,
        county: Optional[Union[list[str], Series[str], str]] = None,
        producing_area: Optional[Union[list[str], Series[str], str]] = None,
        producing_area_id: Optional[Union[list[str], Series[str], str]] = None,
        design_capacity: Optional[str] = None,
        design_capacity_lt: Optional[str] = None,
        design_capacity_lte: Optional[str] = None,
        design_capacity_gt: Optional[str] = None,
        design_capacity_gte: Optional[str] = None,
        uom: Optional[Union[list[str], Series[str], str]] = None,
        operational_capacity: Optional[str] = None,
        operational_capacity_lt: Optional[str] = None,
        operational_capacity_lte: Optional[str] = None,
        operational_capacity_gt: Optional[str] = None,
        operational_capacity_gte: Optional[str] = None,
        scheduled_volume: Optional[str] = None,
        scheduled_volume_lt: Optional[str] = None,
        scheduled_volume_lte: Optional[str] = None,
        scheduled_volume_gt: Optional[str] = None,
        scheduled_volume_gte: Optional[str] = None,
        utilization: Optional[str] = None,
        utilization_lt: Optional[str] = None,
        utilization_lte: Optional[str] = None,
        utilization_gt: Optional[str] = None,
        utilization_gte: Optional[str] = None,
        best_available: Optional[Union[list[bool], Series[bool], bool]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 5000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Daily natural gas pipeline flows in the US lower 48, Canada and Mexico that is accompanied with an array of metadata attributes.

        Parameters
        ----------

         flow_date: Optional[date], optional
             The calendar date or gas day the activity occurred., by default None
         flow_date_gt: Optional[date], optional
             filter by `flow_date > x`, by default None
         flow_date_gte: Optional[date], optional
             filter by `flow_date >= x`, by default None
         flow_date_lt: Optional[date], optional
             filter by `flow_date < x`, by default None
         flow_date_lte: Optional[date], optional
             filter by `flow_date <= x`, by default None
         last_modified_date: Optional[datetime], optional
             Date and time the record was last updated., by default None
         last_modified_date_gt: Optional[datetime], optional
             filter by `last_modified_date > x`, by default None
         last_modified_date_gte: Optional[datetime], optional
             filter by `last_modified_date >= x`, by default None
         last_modified_date_lt: Optional[datetime], optional
             filter by `last_modified_date < x`, by default None
         last_modified_date_lte: Optional[datetime], optional
             filter by `last_modified_date <= x`, by default None
         date_frequency: Optional[Union[list[str], Series[str], str]]
             Daily, Weekly, Monthly, Seasonal, Annual., by default None
         date_frequency_desc: Optional[Union[list[str], Series[str], str]]
             The time period averages of the dataset such as Daily, Weekly, Monthly, Seasonal, Annual. Weekly date frequencies are based on the defined EIA storage week of Friday-Thursday. Seasonal date frequencies define Summer as April to October and Winter as November to March., by default None
         nomination_cycle: Optional[Union[list[str], Series[str], str]]
             Standard NAESB defined nomination cycles for timely (T), evening (E), intraday 1 (I1), intraday 2 (I2) or intraday 3 (I3)., by default None
         legacy_point_logic_point_id: Optional[Union[list[str], Series[str], str]]
             Point ID for a meter used by the legacy PointLogic service., by default None
         legacy_point_logic_lci_id: Optional[Union[list[str], Series[str], str]]
             Alternative point ID for a meter used by the legacy PointLogic service., by default None
         legacy_bentek_point_id: Optional[Union[list[str], Series[str], str]]
             Point ID for a meter used by the legacy Bentek service., by default None
         component_point_id: Optional[Union[list[str], Series[str], str]]
             Point ID for a meter used by the Americas Gas service., by default None
         component_create_date: Optional[datetime], optional
             The date and time stamp of when the record was created., by default None
         component_create_date_gt: Optional[datetime], optional
             filter by `component_create_date > x`, by default None
         component_create_date_gte: Optional[datetime], optional
             filter by `component_create_date >= x`, by default None
         component_create_date_lt: Optional[datetime], optional
             filter by `component_create_date < x`, by default None
         component_create_date_lte: Optional[datetime], optional
             filter by `component_create_date <= x`, by default None
         pipeline_operator_id: Optional[Union[list[str], Series[str], str]]
             ID associated with the common parent owner or operator of a pipeline system., by default None
         pipeline_operator_name: Optional[Union[list[str], Series[str], str]]
             The name of the common parent owner or operator of a pipeline system., by default None
         pipeline_id: Optional[Union[list[str], Series[str], str]]
             The ID given to a pipeline system, utilizes legacy Bentek pipeline ids when applicable., by default None
         pipeline_name: Optional[Union[list[str], Series[str], str]]
             The display name of a pipeline system, utilizes legacy Bentek names when applicable., by default None
         point_name: Optional[Union[list[str], Series[str], str]]
             The display name of a meter or point, utilizes legacy Bentek point name when applicable., by default None
         meter_type_primary: Optional[Union[list[str], Series[str], str]]
             The primary type of classification and purpose of a meter or point, utilizes legacy PointLogic definitions., by default None
         meter_type_id_primary: Optional[Union[list[str], Series[str], str]]
             An ID for the primary type of classification and purpose of a meter or point, utilizes legacy PointLogic definitions and ID., by default None
         meter_type_id_secondary: Optional[Union[list[str], Series[str], str]]
             A secondary type classification and purpose of a meter or point, meant to provide an extra level of detail and utilizes legacy Bentek ids., by default None
         meter_type_secondary: Optional[Union[list[str], Series[str], str]]
             A secondary type classification and purpose of a meter or point, meant to provide an extra level of detail, utilizes legacy Bentek definitions., by default None
         location_type_code: Optional[Union[list[str], Series[str], str]]
             Location type code is a one letter abbreviation of the location description. Location types are sourced from legacy Bentek. These are similar to Flow Direction but serve as a secondary attribute., by default None
         location_description: Optional[Union[list[str], Series[str], str]]
             Location types are sourced from legacy Bentek. These are similar to Flow Direction but serve as a secondary attribute., by default None
         location_type_id: Optional[Union[list[str], Series[str], str]]
             An ID for the location type sourced from legacy Bentek. These are similar to Flow Direction but serve as a secondary attribute., by default None
         flow_direction: Optional[Union[list[str], Series[str], str]]
             Flow direction indicates the orientation of a point such as receipt, delivery, bi-directional or the reported flow direction of segment or compressor. Attribute is sourced from legacy PointLogic. Flow direction is primary to the similar and secondary attribute of Location Type., by default None
         flow_direction_code: Optional[Union[list[str], Series[str], str]]
             A one letter code for Flow Direction such as ‘R’ for receipt or ‘D’ for delivery. Attribute is sourced from legacy PointLogic. Flow direction is primary to the similar and secondary attribute of Location Type., by default None
         flow_direction_id: Optional[Union[list[str], Series[str], str]]
             Flow direction identification number for the orientation of a point such as receipt, delivery, bi-directional or the reported flow direction of segment or compressor. Attribute is sourced from legacy PointLogic. Flow direction is primary to the similar and secondary attribute of Location Type., by default None
         zone: Optional[Union[list[str], Series[str], str]]
             A designation for where on the pipeline system the point is located, corresponding to a pipeline’s operational and market design. Zonal information is sourced from the legacy Bentek service., by default None
         company_id: Optional[Union[list[str], Series[str], str]]
             An ID sourced from the legacy Bentek service used to identify the connecting business or company name of a meter., by default None
         connecting_party: Optional[Union[list[str], Series[str], str]]
             The downstream connecting business name of a meter as reported by the pipeline, utilizes legacy Bentek service., by default None
         loc_prop: Optional[Union[list[str], Series[str], str]]
             The location propriety code reported by the pipeline for a specific meter., by default None
         point_is_active: Optional[Union[list[bool], Series[bool], bool]]
             A true or false return if a point or meter is active and in use within the Americas Gas service., by default None
         domain: Optional[Union[list[str], Series[str], str]]
             US Lower-48, Canada and Mexico., by default None
         domain_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the domain., by default None
         region_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the geographic region., by default None
         region: Optional[Union[list[str], Series[str], str]]
             A defined geographic region within the Americas Gas service. Regions are an aggregation of states or provinces within a country., by default None
         subregion_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the geographic subregion., by default None
         subregion: Optional[Union[list[str], Series[str], str]]
             A defined geographic subregion within the Americas Gas service. A substate geography is sometimes referred to as a subregion. Subregions are an aggregation of specific counties within a region and a country., by default None
         state_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the state or province utilizes legacy Bentek IDs., by default None
         state: Optional[Union[list[str], Series[str], str]]
             The political boundaries that define a state or province within country., by default None
         county_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for the county within a US state in which a meter or point resides, utilizes legacy Bentek IDs., by default None
         county: Optional[Union[list[str], Series[str], str]]
             The political boundaries of a defined county within a US state in which a meter or point resides., by default None
         producing_area: Optional[Union[list[str], Series[str], str]]
             Defined aggregation of counties within a state that is a best fit representation of prominent oil and gas plays and basins., by default None
         producing_area_id: Optional[Union[list[str], Series[str], str]]
             A unique identification number for a defined Producing Area utilizes legacy PointLogic IDs., by default None
         design_capacity: Optional[str], optional
             The volumetric max that a given meter, segment or compressor can receive or deliver as reported by the pipeline., by default None
         design_capacity_gt: Optional[str], optional
             filter by `design_capacity > x`, by default None
         design_capacity_gte: Optional[str], optional
             filter by `design_capacity >= x`, by default None
         design_capacity_lt: Optional[str], optional
             filter by `design_capacity < x`, by default None
         design_capacity_lte: Optional[str], optional
             filter by `design_capacity <= x`, by default None
         uom: Optional[Union[list[str], Series[str], str]]
             Unit of measure., by default None
         operational_capacity: Optional[str], optional
             The pipeline’s reported daily operational capacity for a specific meter, segment or compressor can receive or deliver., by default None
         operational_capacity_gt: Optional[str], optional
             filter by `operational_capacity > x`, by default None
         operational_capacity_gte: Optional[str], optional
             filter by `operational_capacity >= x`, by default None
         operational_capacity_lt: Optional[str], optional
             filter by `operational_capacity < x`, by default None
         operational_capacity_lte: Optional[str], optional
             filter by `operational_capacity <= x`, by default None
         scheduled_volume: Optional[str], optional
             Scheduled volume as reported by the pipeline for a specific meter, segment or compressor., by default None
         scheduled_volume_gt: Optional[str], optional
             filter by `scheduled_volume > x`, by default None
         scheduled_volume_gte: Optional[str], optional
             filter by `scheduled_volume >= x`, by default None
         scheduled_volume_lt: Optional[str], optional
             filter by `scheduled_volume < x`, by default None
         scheduled_volume_lte: Optional[str], optional
             filter by `scheduled_volume <= x`, by default None
         utilization: Optional[str], optional
             Utilization rate in decimal form or in percentage terms. Utilization for a given meter is calculated by dividing the scheduled volume by its operational capacity., by default None
         utilization_gt: Optional[str], optional
             filter by `utilization > x`, by default None
         utilization_gte: Optional[str], optional
             filter by `utilization >= x`, by default None
         utilization_lt: Optional[str], optional
             filter by `utilization < x`, by default None
         utilization_lte: Optional[str], optional
             filter by `utilization <= x`, by default None
         best_available: Optional[Union[list[bool], Series[bool], bool]]
             Best Available refers to the most recent or final nomination cycle for a specific meter on a given flow date. It is indicated as either true or false, with true signifying that the nomination cycle in question is the Best Available. This designation helps in identifying the most current data for operational decisions and reporting purposes., by default None
         filter_exp: Optional[str] = None,
         page: int = 1,
         page_size: int = 5000,
         raw: bool = False,
         paginate: bool = False

        """

        filter_params: List[str] = []
        filter_params.append(list_to_filter("flowDate", flow_date))
        if flow_date_gt is not None:
            filter_params.append(f'flowDate > "{flow_date_gt}"')
        if flow_date_gte is not None:
            filter_params.append(f'flowDate >= "{flow_date_gte}"')
        if flow_date_lt is not None:
            filter_params.append(f'flowDate < "{flow_date_lt}"')
        if flow_date_lte is not None:
            filter_params.append(f'flowDate <= "{flow_date_lte}"')
        filter_params.append(list_to_filter("lastModifiedDate", last_modified_date))
        if last_modified_date_gt is not None:
            filter_params.append(f'lastModifiedDate > "{last_modified_date_gt}"')
        if last_modified_date_gte is not None:
            filter_params.append(f'lastModifiedDate >= "{last_modified_date_gte}"')
        if last_modified_date_lt is not None:
            filter_params.append(f'lastModifiedDate < "{last_modified_date_lt}"')
        if last_modified_date_lte is not None:
            filter_params.append(f'lastModifiedDate <= "{last_modified_date_lte}"')
        filter_params.append(list_to_filter("dateFrequency", date_frequency))
        filter_params.append(list_to_filter("dateFrequencyDesc", date_frequency_desc))
        filter_params.append(list_to_filter("nominationCycle", nomination_cycle))
        filter_params.append(
            list_to_filter("legacyPointLogicPointId", legacy_point_logic_point_id)
        )
        filter_params.append(
            list_to_filter("legacyPointLogicLciId", legacy_point_logic_lci_id)
        )
        filter_params.append(
            list_to_filter("legacyBentekPointId", legacy_bentek_point_id)
        )
        filter_params.append(list_to_filter("componentPointId", component_point_id))
        filter_params.append(
            list_to_filter("componentCreateDate", component_create_date)
        )
        if component_create_date_gt is not None:
            filter_params.append(f'componentCreateDate > "{component_create_date_gt}"')
        if component_create_date_gte is not None:
            filter_params.append(
                f'componentCreateDate >= "{component_create_date_gte}"'
            )
        if component_create_date_lt is not None:
            filter_params.append(f'componentCreateDate < "{component_create_date_lt}"')
        if component_create_date_lte is not None:
            filter_params.append(
                f'componentCreateDate <= "{component_create_date_lte}"'
            )
        filter_params.append(list_to_filter("pipelineOperatorId", pipeline_operator_id))
        filter_params.append(
            list_to_filter("pipelineOperatorName", pipeline_operator_name)
        )
        filter_params.append(list_to_filter("pipelineId", pipeline_id))
        filter_params.append(list_to_filter("pipelineName", pipeline_name))
        filter_params.append(list_to_filter("pointName", point_name))
        filter_params.append(list_to_filter("meterTypePrimary", meter_type_primary))
        filter_params.append(
            list_to_filter("meterTypeIdPrimary", meter_type_id_primary)
        )
        filter_params.append(
            list_to_filter("meterTypeIdSecondary", meter_type_id_secondary)
        )
        filter_params.append(list_to_filter("meterTypeSecondary", meter_type_secondary))
        filter_params.append(list_to_filter("locationTypeCode", location_type_code))
        filter_params.append(
            list_to_filter("locationDescription", location_description)
        )
        filter_params.append(list_to_filter("locationTypeId", location_type_id))
        filter_params.append(list_to_filter("flowDirection", flow_direction))
        filter_params.append(list_to_filter("flowDirectionCode", flow_direction_code))
        filter_params.append(list_to_filter("flowDirectionId", flow_direction_id))
        filter_params.append(list_to_filter("zone", zone))
        filter_params.append(list_to_filter("companyId", company_id))
        filter_params.append(list_to_filter("connectingParty", connecting_party))
        filter_params.append(list_to_filter("locProp", loc_prop))
        filter_params.append(list_to_filter("pointIsActive", point_is_active))
        filter_params.append(list_to_filter("domain", domain))
        filter_params.append(list_to_filter("domainId", domain_id))
        filter_params.append(list_to_filter("regionId", region_id))
        filter_params.append(list_to_filter("region", region))
        filter_params.append(list_to_filter("subregionId", subregion_id))
        filter_params.append(list_to_filter("subregion", subregion))
        filter_params.append(list_to_filter("stateId", state_id))
        filter_params.append(list_to_filter("state", state))
        filter_params.append(list_to_filter("countyId", county_id))
        filter_params.append(list_to_filter("county", county))
        filter_params.append(list_to_filter("producingArea", producing_area))
        filter_params.append(list_to_filter("producingAreaId", producing_area_id))
        filter_params.append(list_to_filter("designCapacity", design_capacity))
        if design_capacity_gt is not None:
            filter_params.append(f'designCapacity > "{design_capacity_gt}"')
        if design_capacity_gte is not None:
            filter_params.append(f'designCapacity >= "{design_capacity_gte}"')
        if design_capacity_lt is not None:
            filter_params.append(f'designCapacity < "{design_capacity_lt}"')
        if design_capacity_lte is not None:
            filter_params.append(f'designCapacity <= "{design_capacity_lte}"')
        filter_params.append(list_to_filter("uom", uom))
        filter_params.append(
            list_to_filter("operationalCapacity", operational_capacity)
        )
        if operational_capacity_gt is not None:
            filter_params.append(f'operationalCapacity > "{operational_capacity_gt}"')
        if operational_capacity_gte is not None:
            filter_params.append(f'operationalCapacity >= "{operational_capacity_gte}"')
        if operational_capacity_lt is not None:
            filter_params.append(f'operationalCapacity < "{operational_capacity_lt}"')
        if operational_capacity_lte is not None:
            filter_params.append(f'operationalCapacity <= "{operational_capacity_lte}"')
        filter_params.append(list_to_filter("scheduledVolume", scheduled_volume))
        if scheduled_volume_gt is not None:
            filter_params.append(f'scheduledVolume > "{scheduled_volume_gt}"')
        if scheduled_volume_gte is not None:
            filter_params.append(f'scheduledVolume >= "{scheduled_volume_gte}"')
        if scheduled_volume_lt is not None:
            filter_params.append(f'scheduledVolume < "{scheduled_volume_lt}"')
        if scheduled_volume_lte is not None:
            filter_params.append(f'scheduledVolume <= "{scheduled_volume_lte}"')
        filter_params.append(list_to_filter("utilization", utilization))
        if utilization_gt is not None:
            filter_params.append(f'utilization > "{utilization_gt}"')
        if utilization_gte is not None:
            filter_params.append(f'utilization >= "{utilization_gte}"')
        if utilization_lt is not None:
            filter_params.append(f'utilization < "{utilization_lt}"')
        if utilization_lte is not None:
            filter_params.append(f'utilization <= "{utilization_lte}"')
        filter_params.append(list_to_filter("bestAvailable", best_available))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path=f"/analytics/gas/na-gas/v1/pipeline-flows",
            params=params,
            df_fn=self._convert_to_df,
            raw=raw,
            paginate=paginate,
        )
        return response


    @staticmethod
    def _convert_to_df(resp: Response) -> pd.DataFrame:
        """
        Converts the API response to a pandas DataFrame and ensures date columns are in ISO8601 format.

        Parameters
        ----------
        resp : Response
            The API response object.

        Returns
        -------
        pd.DataFrame
            The processed DataFrame with date columns in ISO8601 format.
        """
        j = resp.json()
        df = pd.json_normalize(j["results"])

        date_columns = [
            "lastModifiedDate", "flowDate", "forecastDate", "postingDatetime",
            "createDate", "measurementDate", "effectiveDate", "endDate",
            "validFrom", "validTo", "dateEffective", "dateRetire", "dateIssued",
            "date", "contractStartDate", "contractEndDate", "inServiceDate",
            "projectCreatedDate", "projectUpdatedDate", "preFileDate",
            "projectFileDate", "projectApprovalDate"
        ]

        # ISO8601 format
        for column in date_columns:
            if column in df.columns:
                if parse(pd.__version__) >= parse("2"):
                    df[column] = pd.to_datetime(
                        df[column], utc=True, format="ISO8601", errors="coerce"
                    )
                else:
                    df[column] = pd.to_datetime(df[column], errors="coerce", utc=True)

        return df